@@ -18,6 +18,7 @@
 
 import collections
 import contextlib
+import itertools
 import logging
 import re
 
@@ -30,14 +31,13 @@ from sqlalchemy import Column
 from sqlalchemy.engine import Connectable
 from sqlalchemy.engine import reflection
 from sqlalchemy.engine import url as sa_url
-from sqlalchemy.ext.compiler import compiles
+from sqlalchemy import exc
 from sqlalchemy import func
 from sqlalchemy import Index
 from sqlalchemy import inspect
 from sqlalchemy import Integer
 from sqlalchemy import MetaData
 from sqlalchemy.sql.expression import literal_column
-from sqlalchemy.sql.expression import UpdateBase
 from sqlalchemy.sql import text
 from sqlalchemy import String
 from sqlalchemy import Table
@@ -55,30 +55,9 @@ LOG = logging.getLogger(__name__)
 
 _DBURL_REGEX = re.compile(r"[^:]+://([^:]+):([^@]+)@.+")
 
-
-def get_callable_name(function):
-    # TODO(harlowja): Replace this once
-    # it is possible to use https://review.openstack.org/#/c/122495/ which is
-    # a more complete and expansive module that does a similar thing...
-    try:
-        method_self = six.get_method_self(function)
-    except AttributeError:
-        method_self = None
-    if method_self is not None:
-        if isinstance(method_self, six.class_types):
-            im_class = method_self
-        else:
-            im_class = type(method_self)
-        try:
-            parts = (im_class.__module__, function.__qualname__)
-        except AttributeError:
-            parts = (im_class.__module__, im_class.__name__, function.__name__)
-    else:
-        try:
-            parts = (function.__module__, function.__qualname__)
-        except AttributeError:
-            parts = (function.__module__, function.__name__)
-    return '.'.join(parts)
+_VALID_SORT_DIR = [
+    "-".join(x) for x in itertools.product(["asc", "desc"],
+                                           ["nullsfirst", "nullslast"])]
 
 
 def sanitize_db_url(url):
@@ -88,6 +67,71 @@ def sanitize_db_url(url):
     return url
 
 
+def get_unique_keys(model):
+    """Get a list of sets of unique model keys.
+
+    :param model: the ORM model class
+    :rtype: list of sets of strings
+    :return: unique model keys or None if unable to find them
+    """
+
+    try:
+        mapper = inspect(model)
+    except exc.NoInspectionAvailable:
+        return None
+    else:
+        local_table = mapper.local_table
+        base_table = mapper.base_mapper.local_table
+
+        if local_table is None:
+            return None
+
+    # extract result from cache if present
+    has_info = hasattr(local_table, 'info')
+    if has_info:
+        info = local_table.info
+        if 'oslodb_unique_keys' in info:
+            return info['oslodb_unique_keys']
+
+    res = []
+    try:
+        constraints = base_table.constraints
+    except AttributeError:
+        constraints = []
+    for constraint in constraints:
+        # filter out any CheckConstraints
+        if isinstance(constraint, (sqlalchemy.UniqueConstraint,
+                                   sqlalchemy.PrimaryKeyConstraint)):
+            res.append({c.name for c in constraint.columns})
+    try:
+        indexes = base_table.indexes
+    except AttributeError:
+        indexes = []
+    for index in indexes:
+        if index.unique:
+            res.append({c.name for c in index.columns})
+    # cache result for next calls with the same model
+    if has_info:
+        info['oslodb_unique_keys'] = res
+    return res
+
+
+def _stable_sorting_order(model, sort_keys):
+    """Check whetever the sorting order is stable.
+
+    :return: True if it is stable, False if it's not, None if it's impossible
+    to determine.
+    """
+    keys = get_unique_keys(model)
+    if keys is None:
+        return None
+    sort_keys_set = set(sort_keys)
+    for unique_keys in keys:
+        if unique_keys.issubset(sort_keys_set):
+            return True
+    return False
+
+
 # copy from glance/db/sqlalchemy/api.py
 def paginate_query(query, model, limit, sort_keys, marker=None,
                    sort_dir=None, sort_dirs=None):
@@ -115,16 +159,16 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     :param marker: the last item of the previous page; we returns the next
                     results after this value.
     :param sort_dir: direction in which results should be sorted (asc, desc)
+                     suffix -nullsfirst, -nullslast can be added to defined
+                     the ordering of null values
     :param sort_dirs: per-column array of sort_dirs, corresponding to sort_keys
 
     :rtype: sqlalchemy.orm.query.Query
     :return: The query with sorting/pagination added.
     """
-
-    if 'id' not in sort_keys:
-        # TODO(justinsb): If this ever gives a false-positive, check
-        # the actual primary key, rather than assuming its id
-        LOG.warning(_LW('Id not in sort_keys; is sort_keys unique?'))
+    if _stable_sorting_order(model, sort_keys) is False:
+        LOG.warning(_LW('Unique keys not in sort_keys. '
+                        'The sorting order may be unstable.'))
 
     assert(not (sort_dir and sort_dirs))
 
@@ -141,18 +185,31 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     # Add sorting
     for current_sort_key, current_sort_dir in zip(sort_keys, sort_dirs):
         try:
+            inspect(model).all_orm_descriptors[current_sort_key]
+        except KeyError:
+            raise exception.InvalidSortKey(current_sort_key)
+        else:
+            sort_key_attr = getattr(model, current_sort_key)
+
+        try:
+            main_sort_dir, __, null_sort_dir = current_sort_dir.partition("-")
             sort_dir_func = {
                 'asc': sqlalchemy.asc,
                 'desc': sqlalchemy.desc,
-            }[current_sort_dir]
+            }[main_sort_dir]
+
+            null_order_by_stmt = {
+                "": None,
+                "nullsfirst": sort_key_attr.is_(None),
+                "nullslast": sort_key_attr.isnot(None),
+            }[null_sort_dir]
         except KeyError:
             raise ValueError(_("Unknown sort direction, "
-                               "must be 'desc' or 'asc'"))
-        try:
-            sort_key_attr = inspect(model).\
-                all_orm_descriptors[current_sort_key]
-        except KeyError:
-            raise exception.InvalidSortKey()
+                               "must be one of: %s") %
+                             ", ".join(_VALID_SORT_DIR))
+
+        if null_order_by_stmt is not None:
+            query = query.order_by(sqlalchemy.desc(null_order_by_stmt))
         query = query.order_by(sort_dir_func(sort_key_attr))
 
     # Add pagination
@@ -171,7 +228,7 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
                 crit_attrs.append((model_attr == marker_values[j]))
 
             model_attr = getattr(model, sort_keys[i])
-            if sort_dirs[i] == 'desc':
+            if sort_dirs[i].startswith('desc'):
                 crit_attrs.append((model_attr < marker_values[i]))
             else:
                 crit_attrs.append((model_attr > marker_values[i]))
@@ -188,6 +245,18 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     return query
 
 
+def to_list(x, default=None):
+    if x is None:
+        return default
+    if not isinstance(x, collections.Iterable) or \
+            isinstance(x, six.string_types):
+        return [x]
+    elif isinstance(x, list):
+        return x
+    else:
+        return list(x)
+
+
 def _read_deleted_filter(query, db_model, deleted):
     if 'deleted' not in db_model.__table__.columns:
         raise ValueError(_("There is no `deleted` column in `%s` table. "
@@ -346,19 +415,40 @@ def get_table(engine, name):
     return Table(name, metadata, autoload=True)
 
 
-class InsertFromSelect(UpdateBase):
-    """Form the base for `INSERT INTO table (SELECT ... )` statement."""
-    def __init__(self, table, select):
-        self.table = table
-        self.select = select
+class InsertFromSelect(object):
+    """Form the base for `INSERT INTO table (SELECT ... )` statement.
 
+    DEPRECATED: this class is deprecated and will be removed from oslo_db
+    in a few releases. Use default SQLAlchemy insert from select implementation
+    instead
 
-@compiles(InsertFromSelect)
-def visit_insert_from_select(element, compiler, **kw):
-    """Form the `INSERT INTO table (SELECT ... )` statement."""
-    return "INSERT INTO %s %s" % (
-        compiler.process(element.table, asfrom=True),
-        compiler.process(element.select))
+    :param table: table to insert records
+    :param select: select query
+    :param cols: list of columns to specify in insert clause
+    :return: SQLAlchemy :class:`Insert` object instance
+
+    Usage:
+
+    .. code-block:: python
+
+      select = sql.select(table_from)
+      insert = InsertFromSelect(table_to, select,
+                                ['id', 'name', 'insert_date'])
+      engine.execute(insert)
+
+    """
+    # NOTE(tdurakov): Insert from select implementation added to SQLAlchemy
+    # starting from version 0.8.7. Default SQLAlchemy implementation should be
+    # used instead of this. Deprecated.
+
+    def __new__(cls, table, select, cols=None):
+        if not cols:
+            cols = [c.name for c in table.c]
+
+        return table.insert(inline=True).from_select(cols, select)
+
+    def __init__(self, table, select, cols=None):
+        pass
 
 
 def _get_not_supported_column(col_name_col_instance, column_name):
@@ -563,6 +653,19 @@ def change_deleted_column_type_to_id_type(migrate_engine, table_name,
     _restore_indexes_on_deleted_columns(migrate_engine, table_name, indexes)
 
 
+def _is_deleted_column_constraint(constraint):
+    # NOTE(boris-42): There is no other way to check is CheckConstraint
+    #                 associated with deleted column.
+    if not isinstance(constraint, CheckConstraint):
+        return False
+    sqltext = str(constraint.sqltext)
+    # NOTE(zzzeek): SQLite never reflected CHECK contraints here
+    # in any case until version 1.1.   Safe to assume that any CHECK
+    # that's talking about the value of "deleted in (something)" is
+    # the boolean constraint we're looking to get rid of.
+    return bool(re.match(r".*deleted in \(.*\)", sqltext, re.I))
+
+
 def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
                                                   **col_name_col_instance):
     # NOTE(boris-42): sqlaclhemy-migrate can't drop column with check
@@ -593,25 +696,9 @@ def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
                                  default=default_deleted_value)
         columns.append(column_copy)
 
-    def is_deleted_column_constraint(constraint):
-        # NOTE(boris-42): There is no other way to check is CheckConstraint
-        #                 associated with deleted column.
-        if not isinstance(constraint, CheckConstraint):
-            return False
-        sqltext = str(constraint.sqltext)
-        # NOTE(I159): in order to omit the CHECK constraint corresponding
-        # to `deleted` column we have to test these patterns which may
-        # vary depending on the SQLAlchemy version used.
-        constraint_markers = (
-            "deleted in (0, 1)",
-            "deleted IN (:deleted_1, :deleted_2)",
-            "deleted IN (:param_1, :param_2)"
-        )
-        return any(sqltext.endswith(marker) for marker in constraint_markers)
-
     constraints = []
     for constraint in table.constraints:
-        if not is_deleted_column_constraint(constraint):
+        if not _is_deleted_column_constraint(constraint):
             constraints.append(constraint.copy())
 
     new_table = Table(table_name + "__tmp__", meta,
@@ -653,7 +740,7 @@ def get_connect_string(backend, database, user=None, passwd=None,
     Try to get a connection with a very specific set of values, if we get
     these then we'll run the tests, otherwise they are skipped
 
-    DEPRECATED: this function is deprecated and will be removed from oslo.db
+    DEPRECATED: this function is deprecated and will be removed from oslo_db
     in a few releases. Please use the provisioning system for dealing
     with URLs and database provisioning.
 
@@ -674,7 +761,7 @@ def is_backend_avail(backend, database, user=None, passwd=None):
     """Return True if the given backend is available.
 
 
-    DEPRECATED: this function is deprecated and will be removed from oslo.db
+    DEPRECATED: this function is deprecated and will be removed from oslo_db
     in a few releases. Please use the provisioning system to access
     databases based on backend availability.
 
