@@ -1,3 +1,5 @@
+#    -*- encoding: utf-8 -*-
+#
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
 #    a copy of the License at
@@ -19,12 +21,15 @@ import mock
 from oslotest import base as oslo_test_base
 import six
 import sqlalchemy as sqla
+from sqlalchemy.engine import url as sqla_url
+from sqlalchemy import event
+import sqlalchemy.exc
+from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.orm import mapper
 
 from oslo_db import exception
-from oslo_db.sqlalchemy import compat
+from oslo_db.sqlalchemy import engines
 from oslo_db.sqlalchemy import exc_filters
-from oslo_db.sqlalchemy import session
 from oslo_db.sqlalchemy import test_base
 from oslo_db.tests import utils as test_utils
 
@@ -45,12 +50,26 @@ class _SQLAExceptionMatcher(object):
         if isinstance(exception_type, (type, tuple)):
             self.assertTrue(issubclass(exc.__class__, exception_type))
         else:
-            self.assertEqual(exc.__class__.__name__, exception_type)
-        self.assertEqual(str(exc.orig).lower(), message.lower())
+            self.assertEqual(exception_type, exc.__class__.__name__)
+        if isinstance(message, tuple):
+            self.assertEqual(
+                [m.lower()
+                 if isinstance(m, six.string_types) else m for m in message],
+                [a.lower()
+                 if isinstance(a, six.string_types) else a
+                 for a in exc.orig.args]
+            )
+        else:
+            self.assertEqual(message.lower(), str(exc.orig).lower())
         if sql is not None:
-            self.assertEqual(exc.statement, sql)
-        if params is not None:
-            self.assertEqual(exc.params, params)
+            if params is not None:
+                if '?' in exc.statement:
+                    self.assertEqual(sql, exc.statement)
+                    self.assertEqual(params, exc.params)
+                else:
+                    self.assertEqual(sql % params, exc.statement % exc.params)
+            else:
+                self.assertEqual(sql, exc.statement)
 
 
 class TestsExceptionFilter(_SQLAExceptionMatcher, oslo_test_base.BaseTestCase):
@@ -63,6 +82,9 @@ class TestsExceptionFilter(_SQLAExceptionMatcher, oslo_test_base.BaseTestCase):
 
         """
 
+    class DataError(Error):
+        pass
+
     class OperationalError(Error):
         pass
 
@@ -226,6 +248,201 @@ class TestFallthroughsAndNonDBAPI(TestsExceptionFilter):
         self.assertEqual("mysqldb has an attribute error", matched.args[0])
 
 
+class TestNonExistentConstraint(
+        _SQLAExceptionMatcher,
+        test_base.DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentConstraint, self).setUp()
+
+        meta = sqla.MetaData(bind=self.engine)
+
+        self.table_1 = sqla.Table(
+            "resource_foo", meta,
+            sqla.Column("id", sqla.Integer, primary_key=True),
+            mysql_engine='InnoDB',
+            mysql_charset='utf8',
+        )
+        self.table_1.create()
+
+
+class TestNonExistentConstraintPostgreSQL(
+        TestNonExistentConstraint,
+        test_base.PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentConstraint,
+            self.engine.execute,
+            sqla.schema.DropConstraint(
+                sqla.ForeignKeyConstraint(["id"], ["baz.id"],
+                                          name="bar_fkey",
+                                          table=self.table_1)),
+        )
+        self.assertInnerException(
+            matched,
+            "ProgrammingError",
+            "constraint \"bar_fkey\" of relation "
+            "\"resource_foo\" does not exist\n",
+            "ALTER TABLE resource_foo DROP CONSTRAINT bar_fkey",
+        )
+        self.assertEqual("resource_foo", matched.table)
+        self.assertEqual("bar_fkey", matched.constraint)
+
+
+class TestNonExistentConstraintMySQL(
+        TestNonExistentConstraint,
+        test_base.MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentConstraint,
+            self.engine.execute,
+            sqla.schema.DropConstraint(
+                sqla.ForeignKeyConstraint(["id"], ["baz.id"],
+                                          name="bar_fkey",
+                                          table=self.table_1)),
+        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.InternalError)
+        if matched.table is not None:
+            self.assertEqual("resource_foo", matched.table)
+        if matched.constraint is not None:
+            self.assertEqual("bar_fkey", matched.constraint)
+
+
+class TestNonExistentTable(
+        _SQLAExceptionMatcher,
+        test_base.DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentTable, self).setUp()
+
+        self.meta = sqla.MetaData(bind=self.engine)
+
+        self.table_1 = sqla.Table(
+            "foo", self.meta,
+            sqla.Column("id", sqla.Integer, primary_key=True),
+            mysql_engine='InnoDB',
+            mysql_charset='utf8',
+        )
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        self.assertInnerException(
+            matched,
+            "OperationalError",
+            "no such table: foo",
+            "\nDROP TABLE foo",
+        )
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentTablePostgreSQL(
+        TestNonExistentTable,
+        test_base.PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        self.assertInnerException(
+            matched,
+            "ProgrammingError",
+            "table \"foo\" does not exist\n",
+            "\nDROP TABLE foo",
+        )
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentTableMySQL(
+        TestNonExistentTable,
+        test_base.MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.InternalError)
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentDatabase(
+        _SQLAExceptionMatcher,
+        test_base.DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentDatabase, self).setUp()
+
+        url = sqla_url.make_url(str(self.engine.url))
+        url.database = 'non_existent_database'
+        self.url = url
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            sqla_url.make_url(
+                'sqlite:////non_existent_dir/non_existent_database')
+        )
+        self.assertIsNone(matched.database)
+        self.assertInnerException(
+            matched,
+            sqlalchemy.exc.OperationalError,
+            'unable to open database file',
+        )
+
+
+class TestNonExistentDatabaseMySQL(
+        TestNonExistentDatabase,
+        test_base.MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            self.url
+        )
+        self.assertEqual('non_existent_database', matched.database)
+        # NOTE(rpodolyaka) cannot check precisely with assertInnerException
+        # since MySQL errors are not the same depending on its version
+        self.assertIsInstance(
+            matched.inner_exception,
+            (sqlalchemy.exc.InternalError, sqlalchemy.exc.OperationalError),
+        )
+
+
+class TestNonExistentDatabasePostgreSQL(
+        TestNonExistentDatabase,
+        test_base.PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            self.url
+        )
+        self.assertEqual('non_existent_database', matched.database)
+        self.assertInnerException(
+            matched,
+            sqlalchemy.exc.OperationalError,
+            'fatal:  database "non_existent_database" does not exist\n',
+        )
+
+
 class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
 
     def setUp(self):
@@ -233,14 +450,14 @@ class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
 
         meta = sqla.MetaData(bind=self.engine)
 
-        table_1 = sqla.Table(
+        self.table_1 = sqla.Table(
             "resource_foo", meta,
             sqla.Column("id", sqla.Integer, primary_key=True),
             sqla.Column("foo", sqla.Integer),
             mysql_engine='InnoDB',
             mysql_charset='utf8',
         )
-        table_1.create()
+        self.table_1.create()
 
         self.table_2 = sqla.Table(
             "resource_entity", meta,
@@ -274,6 +491,30 @@ class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
         self.assertIsNone(matched.key)
         self.assertIsNone(matched.key_table)
 
+    def test_raise_delete(self):
+        self.engine.execute("PRAGMA foreign_keys = ON;")
+
+        with self.engine.connect() as conn:
+            conn.execute(self.table_1.insert({"id": 1234, "foo": 42}))
+            conn.execute(self.table_2.insert({"id": 4321, "foo_id": 1234}))
+        matched = self.assertRaises(
+            exception.DBReferenceError,
+            self.engine.execute,
+            self.table_1.delete()
+        )
+        self.assertInnerException(
+            matched,
+            "IntegrityError",
+            "foreign key constraint failed",
+            "DELETE FROM resource_foo",
+            (),
+        )
+
+        self.assertIsNone(matched.table)
+        self.assertIsNone(matched.constraint)
+        self.assertIsNone(matched.key)
+        self.assertIsNone(matched.key_table)
+
 
 class TestReferenceErrorPostgreSQL(TestReferenceErrorSQLite,
                                    test_base.PostgreSQLOpportunisticTestCase):
@@ -300,6 +541,31 @@ class TestReferenceErrorPostgreSQL(TestReferenceErrorSQLite,
         self.assertEqual("foo_id", matched.key)
         self.assertEqual("resource_foo", matched.key_table)
 
+    def test_raise_delete(self):
+        with self.engine.connect() as conn:
+            conn.execute(self.table_1.insert({"id": 1234, "foo": 42}))
+            conn.execute(self.table_2.insert({"id": 4321, "foo_id": 1234}))
+        matched = self.assertRaises(
+            exception.DBReferenceError,
+            self.engine.execute,
+            self.table_1.delete()
+        )
+        self.assertInnerException(
+            matched,
+            "IntegrityError",
+            "update or delete on table \"resource_foo\" violates foreign key "
+            "constraint \"foo_fkey\" on table \"resource_entity\"\n"
+            "DETAIL:  Key (id)=(1234) is still referenced from "
+            "table \"resource_entity\".\n",
+            "DELETE FROM resource_foo",
+            {},
+        )
+
+        self.assertEqual("resource_foo", matched.table)
+        self.assertEqual("foo_fkey", matched.constraint)
+        self.assertEqual("id", matched.key)
+        self.assertEqual("resource_entity", matched.key_table)
+
 
 class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
                               test_base.MySQLOpportunisticTestCase):
@@ -310,16 +576,11 @@ class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
             self.table_2.insert({'id': 1, 'foo_id': 2})
         )
 
-        self.assertInnerException(
-            matched,
-            "IntegrityError",
-            "(1452, 'Cannot add or update a child row: a "
-            "foreign key constraint fails (`{0}`.`resource_entity`, "
-            "CONSTRAINT `foo_fkey` FOREIGN KEY (`foo_id`) REFERENCES "
-            "`resource_foo` (`id`))')".format(self.engine.url.database),
-            "INSERT INTO resource_entity (id, foo_id) VALUES (%s, %s)",
-            (1, 2)
-        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(matched.inner_exception.orig.args[0], 1452)
         self.assertEqual("resource_entity", matched.table)
         self.assertEqual("foo_fkey", matched.constraint)
         self.assertEqual("foo_id", matched.key)
@@ -338,21 +599,151 @@ class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
                 self.table_2.insert({'id': 1, 'foo_id': 2})
             )
 
-        self.assertInnerException(
-            matched,
-            "IntegrityError",
-            '(1452, \'Cannot add or update a child row: a '
-            'foreign key constraint fails ("{0}"."resource_entity", '
-            'CONSTRAINT "foo_fkey" FOREIGN KEY ("foo_id") REFERENCES '
-            '"resource_foo" ("id"))\')'.format(self.engine.url.database),
-            "INSERT INTO resource_entity (id, foo_id) VALUES (%s, %s)",
-            (1, 2)
-        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(matched.inner_exception.orig.args[0], 1452)
         self.assertEqual("resource_entity", matched.table)
         self.assertEqual("foo_fkey", matched.constraint)
         self.assertEqual("foo_id", matched.key)
         self.assertEqual("resource_foo", matched.key_table)
 
+    def test_raise_delete(self):
+        with self.engine.connect() as conn:
+            conn.execute(self.table_1.insert({"id": 1234, "foo": 42}))
+            conn.execute(self.table_2.insert({"id": 4321, "foo_id": 1234}))
+        matched = self.assertRaises(
+            exception.DBReferenceError,
+            self.engine.execute,
+            self.table_1.delete()
+        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(1451, matched.inner_exception.orig.args[0])
+        self.assertEqual("resource_entity", matched.table)
+        self.assertEqual("foo_fkey", matched.constraint)
+        self.assertEqual("foo_id", matched.key)
+        self.assertEqual("resource_foo", matched.key_table)
+
+
+class TestExceptionCauseMySQLSavepoint(test_base.MySQLOpportunisticTestCase):
+    def setUp(self):
+        super(TestExceptionCauseMySQLSavepoint, self).setUp()
+
+        Base = declarative_base()
+
+        class A(Base):
+            __tablename__ = 'a'
+
+            id = sqla.Column(sqla.Integer, primary_key=True)
+
+            __table_args__ = {'mysql_engine': 'InnoDB'}
+
+        Base.metadata.create_all(self.engine)
+
+        self.A = A
+
+    def test_cause_for_failed_flush_plus_no_savepoint(self):
+        session = self.sessionmaker()
+
+        with session.begin():
+            session.add(self.A(id=1))
+        try:
+
+            with session.begin():
+
+                try:
+                    with session.begin_nested():
+                        session.execute("rollback")
+                        session.add(self.A(id=1))
+
+                # outermost is the failed SAVEPOINT rollback
+                # from the "with session.begin_nested()"
+                except exception.DBError as dbe_inner:
+
+                    # first "cause" is the failed SAVEPOINT rollback
+                    # from inside of flush(), when it fails
+                    self.assertTrue(
+                        isinstance(
+                            dbe_inner.cause,
+                            exception.DBError
+                        )
+                    )
+
+                    # second "cause" is then the actual DB duplicate
+                    self.assertTrue(
+                        isinstance(
+                            dbe_inner.cause.cause,
+                            exception.DBDuplicateEntry
+                        )
+                    )
+        except exception.DBError as dbe_outer:
+            self.assertTrue(
+                isinstance(
+                    dbe_outer.cause,
+                    exception.DBDuplicateEntry
+                )
+            )
+
+        # resets itself afterwards
+        try:
+            with session.begin():
+                session.add(self.A(id=1))
+        except exception.DBError as dbe_outer:
+            self.assertIsNone(dbe_outer.cause)
+
+
+class TestDBDataErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
+
+    def setUp(self):
+        super(TestDBDataErrorSQLite, self).setUp()
+
+        if six.PY3:
+            self.skip("SQLite database supports unicode value for python3")
+
+        meta = sqla.MetaData(bind=self.engine)
+
+        self.table_1 = sqla.Table(
+            "resource_foo", meta,
+            sqla.Column("name", sqla.String),
+        )
+        self.table_1.create()
+
+    def test_raise(self):
+
+        matched = self.assertRaises(
+            exception.DBDataError,
+            self.engine.execute,
+            self.table_1.insert({'name': u'\u2713'.encode('utf-8')})
+        )
+
+        self.assertInnerException(
+            matched,
+            "ProgrammingError",
+            "You must not use 8-bit bytestrings unless you use a "
+            "text_factory that can interpret 8-bit bytestrings "
+            "(like text_factory = str). It is highly recommended that "
+            "you instead just switch your application to Unicode strings.",
+            "INSERT INTO resource_foo (name) VALUES (?)",
+            (u'\u2713'.encode('utf-8'),)
+        )
+
+
+class TestConstraint(TestsExceptionFilter):
+    def test_postgresql(self):
+        matched = self._run_test(
+            "postgresql", "insert into resource some_values",
+            self.IntegrityError(
+                "new row for relation \"resource\" violates "
+                "check constraint \"ck_started_before_ended\""),
+            exception.DBConstraintError,
+        )
+        self.assertEqual("resource", matched.table)
+        self.assertEqual("ck_started_before_ended", matched.check_name)
+
 
 class TestDuplicate(TestsExceptionFilter):
 
@@ -395,11 +786,15 @@ class TestDuplicate(TestsExceptionFilter):
             "PRIMARY KEY must be unique 'insert into t values(10)'",
             expected_columns=[])
 
-    def test_mysql_mysqldb(self):
+    def test_mysql_pymysql(self):
         self._run_dupe_constraint_test(
             "mysql",
             '(1062, "Duplicate entry '
             '\'2-3\' for key \'uniq_tbl0a0b\'")', expected_value='2-3')
+        self._run_dupe_constraint_test(
+            "mysql",
+            '(1062, "Duplicate entry '
+            '\'\' for key \'uniq_tbl0a0b\'")', expected_value='')
 
     def test_mysql_mysqlconnector(self):
         self._run_dupe_constraint_test(
@@ -424,6 +819,24 @@ class TestDuplicate(TestsExceptionFilter):
             expected_value='2'
         )
 
+    def test_mysql_binary(self):
+        self._run_dupe_constraint_test(
+            "mysql",
+            "(1062, \'Duplicate entry "
+            "\\\'\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E\\\' "
+            "for key \\\'PRIMARY\\\'\')",
+            expected_columns=['PRIMARY'],
+            expected_value="\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E"
+        )
+        self._run_dupe_constraint_test(
+            "mysql",
+            "(1062, \'Duplicate entry "
+            "''\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E!,' "
+            "for key 'PRIMARY'\')",
+            expected_columns=['PRIMARY'],
+            expected_value="'\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E!,"
+        )
+
     def test_postgresql_single(self):
         self._run_dupe_constraint_test(
             'postgresql',
@@ -498,9 +911,9 @@ class TestDeadlock(TestsExceptionFilter):
         if isinstance(matched, exception.DBError):
             matched = matched.inner_exception
 
-        self.assertEqual(matched.orig.__class__.__name__, expected_dbapi_cls)
+        self.assertEqual(expected_dbapi_cls, matched.orig.__class__.__name__)
 
-    def test_mysql_mysqldb_deadlock(self):
+    def test_mysql_pymysql_deadlock(self):
         self._run_deadlock_detect_test(
             "mysql",
             "(1213, 'Deadlock found when trying "
@@ -508,11 +921,12 @@ class TestDeadlock(TestsExceptionFilter):
             "transaction')"
         )
 
-    def test_mysql_mysqldb_galera_deadlock(self):
+    def test_mysql_pymysql_galera_deadlock(self):
         self._run_deadlock_detect_test(
             "mysql",
             "(1205, 'Lock wait timeout exceeded; "
-            "try restarting transaction')"
+            "try restarting transaction')",
+            orig_exception_cls=self.InternalError
         )
 
     def test_mysql_mysqlconnector_deadlock(self):
@@ -568,6 +982,34 @@ class TestDeadlock(TestsExceptionFilter):
         )
 
 
+class TestDataError(TestsExceptionFilter):
+    def _run_bad_data_test(self, dialect_name, message, error_class):
+        self._run_test(dialect_name,
+                       "INSERT INTO TABLE some_values",
+                       error_class(message),
+                       exception.DBDataError)
+
+    def test_bad_data_incorrect_string(self):
+        # Error sourced from https://bugs.launchpad.net/cinder/+bug/1393871
+        self._run_bad_data_test("mysql",
+                                '(1366, "Incorrect string value: \'\\xF0\' '
+                                'for column \'resource\' at row 1"',
+                                self.OperationalError)
+
+    def test_bad_data_out_of_range(self):
+        # Error sourced from https://bugs.launchpad.net/cinder/+bug/1463379
+        self._run_bad_data_test("mysql",
+                                '(1264, "Out of range value for column '
+                                '\'resource\' at row 1"',
+                                self.DataError)
+
+    def test_data_too_long_for_column(self):
+        self._run_bad_data_test("mysql",
+                                '(1406, "Data too long for column '
+                                '\'resource\' at row 1"',
+                                self.DataError)
+
+
 class IntegrationTest(test_base.DbTestCase):
     """Test an actual error-raising round trips against the database."""
 
@@ -678,7 +1120,7 @@ class TestDBDisconnected(TestsExceptionFilter):
             dialect_name, exception, num_disconnects, is_disconnect=True):
         engine = self.engine
 
-        compat.engine_connect(engine, session._connect_ping_listener)
+        event.listen(engine, "engine_connect", engines._connect_ping_listener)
 
         real_do_execute = engine.dialect.do_execute
         counter = itertools.count(1)
@@ -705,7 +1147,7 @@ class TestDBDisconnected(TestsExceptionFilter):
         with self._fixture(dialect_name, exc_obj, 1, is_disconnect):
             conn = self.engine.connect()
             with conn.begin():
-                self.assertEqual(conn.scalar(sqla.select([1])), 1)
+                self.assertEqual(1, conn.scalar(sqla.select([1])))
                 self.assertFalse(conn.closed)
                 self.assertFalse(conn.invalidated)
                 self.assertTrue(conn.in_transaction())
@@ -718,7 +1160,7 @@ class TestDBDisconnected(TestsExceptionFilter):
 
         # test implicit execution
         with self._fixture(dialect_name, exc_obj, 1):
-            self.assertEqual(self.engine.scalar(sqla.select([1])), 1)
+            self.assertEqual(1, self.engine.scalar(sqla.select([1])))
 
     def test_mysql_ping_listener_disconnected(self):
         for code in [2006, 2013, 2014, 2045, 2055]:
@@ -738,6 +1180,24 @@ class TestDBDisconnected(TestsExceptionFilter):
                 is_disconnect=False
             )
 
+    def test_mysql_galera_non_primary_disconnected(self):
+        self._test_ping_listener_disconnected(
+            "mysql",
+            self.OperationalError('(1047, \'Unknown command\') '
+                                  '\'SELECT DATABASE()\' ()')
+        )
+
+    def test_mysql_galera_non_primary_disconnected_regex_only(self):
+        # intentionally set the is_disconnect flag to False
+        # in the "sqlalchemy" layer to make sure the regexp
+        # on _is_db_connection_error is catching
+        self._test_ping_listener_disconnected(
+            "mysql",
+            self.OperationalError('(1047, \'Unknown command\') '
+                                  '\'SELECT DATABASE()\' ()'),
+            is_disconnect=False
+        )
+
     def test_db2_ping_listener_disconnected(self):
         self._test_ping_listener_disconnected(
             "ibm_db_sa",
@@ -789,7 +1249,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
 
         with self._dbapi_fixture(dialect_name):
             with mock.patch.object(engine.dialect, "connect", cant_connect):
-                return session._test_connection(engine, retries, .01)
+                return engines._test_connection(engine, retries, .01)
 
     def test_connect_no_retries(self):
         conn = self._run_test(
@@ -807,7 +1267,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, -1
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_connect_retry_past_failure(self):
         conn = self._run_test(
@@ -816,7 +1276,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, 3
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_connect_retry_not_candidate_exception(self):
         self.assertRaises(
@@ -844,7 +1304,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, -1
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_db2_error_negative(self):
         self.assertRaises(
@@ -860,7 +1320,8 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
 
     def setUp(self):
         super(TestDBConnectPingWrapping, self).setUp()
-        compat.engine_connect(self.engine, session._connect_ping_listener)
+        event.listen(
+            self.engine, "engine_connect", engines._connect_ping_listener)
 
     @contextlib.contextmanager
     def _fixture(
@@ -900,7 +1361,7 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
             self, dialect_name, exc_obj, is_disconnect=True):
         with self._fixture(dialect_name, exc_obj, 3, is_disconnect):
             conn = self.engine.connect()
-            self.assertEqual(conn.scalar(sqla.select([1])), 1)
+            self.assertEqual(1, conn.scalar(sqla.select([1])))
             conn.close()
 
         with self._fixture(dialect_name, exc_obj, 1, is_disconnect):
