@@ -24,15 +24,21 @@ from six.moves.urllib import parse
 import sqlalchemy
 from sqlalchemy.dialects import mysql
 from sqlalchemy import Boolean, Index, Integer, DateTime, String, SmallInteger
+from sqlalchemy import CheckConstraint
 from sqlalchemy import MetaData, Table, Column, ForeignKey
 from sqlalchemy.engine import reflection
 from sqlalchemy.engine import url as sa_url
 from sqlalchemy.exc import OperationalError
 from sqlalchemy.ext.declarative import declarative_base
+from sqlalchemy.ext.hybrid import hybrid_property
+from sqlalchemy.orm import mapper
+from sqlalchemy.orm import Session
 from sqlalchemy.sql import select
 from sqlalchemy.types import UserDefinedType, NullType
+from sqlalchemy.dialects.postgresql import psycopg2
 
 from oslo_db import exception
+from oslo_db.sqlalchemy.compat import utils as compat_utils
 from oslo_db.sqlalchemy import models
 from oslo_db.sqlalchemy import provision
 from oslo_db.sqlalchemy import session
@@ -42,7 +48,7 @@ from oslo_db.tests import utils as test_utils
 
 
 Base = declarative_base()
-SA_VERSION = tuple(map(int, sqlalchemy.__version__.split('.')))
+SA_VERSION = compat_utils.SQLA_VERSION
 
 
 class TestSanitizeDbUrl(test_base.BaseTestCase):
@@ -72,10 +78,68 @@ class FakeTable(Base):
     project_id = Column(String(50))
     snapshot_id = Column(String(50))
 
+    # mox is comparing in some awkward way that
+    # in this case requires the same identity of object
+    _expr_to_appease_mox = project_id + snapshot_id
+
+    @hybrid_property
+    def some_hybrid(self):
+        raise NotImplementedError()
+
+    @some_hybrid.expression
+    def some_hybrid(cls):
+        return cls._expr_to_appease_mox
+
     def foo(self):
         pass
 
 
+class FakeTableJoinedInh(FakeTable):
+    __tablename__ = 'fake_table_inh'
+
+    id = Column(String(50), ForeignKey('fake_table.user_id'))
+
+
+class FakeTableSingleInh(FakeTable):
+    __mapper_args__ = {'polymorphic_identity': 'foo'}
+
+
+class FakeTableWithMultipleKeys(Base):
+    __tablename__ = 'fake_table_multiple_keys'
+
+    key1 = Column(String(50), primary_key=True)
+    key2 = Column(String(50), primary_key=True)
+    key3 = Column(String(50))
+
+
+class FakeTableWithIndexes(Base):
+    __tablename__ = 'fake_table_unique_index'
+
+    id = Column(String(50), primary_key=True)
+    key1 = Column(String(50))
+    key2 = Column(String(50))
+    key3 = Column(String(50))
+
+    __table_args__ = (
+        Index('idx_unique', 'key1', 'key2', unique=True),
+        Index('idx_unique', 'key1', 'key3', unique=False),
+    )
+
+
+class FakeTableClassicalyMapped(object):
+    pass
+
+
+fake_table = Table(
+    'fake_table_classically_mapped',
+    Base.metadata,
+    Column('id', Integer, primary_key=True),
+    Column('key', String(50))
+)
+
+mapper(FakeTableClassicalyMapped, fake_table)
+
+
 class FakeModel(object):
     def __init__(self, values):
         self.values = values
@@ -133,6 +197,23 @@ class TestPaginateQuery(test_base.BaseTestCase):
                              ['user_id', 'project_id'],
                              sort_dirs=['asc', 'desc'])
 
+    def test_invalid_sort_key_str(self):
+        self.assertEqual("Sort key supplied is invalid: None",
+                         str(exception.InvalidSortKey()))
+        self.assertEqual("Sort key supplied is invalid: lol",
+                         str(exception.InvalidSortKey("lol")))
+        self.assertEqual("Sort key supplied is invalid: lol",
+                         exception.InvalidSortKey("lol").message)
+
+    def test_invalid_unicode_paramater_str(self):
+        self.assertEqual(
+            "Invalid Parameter: Encoding directive wasn't provided.",
+            str(exception.DBInvalidUnicodeParameter()))
+        self.assertEqual(
+            "Invalid Parameter: Encoding directive wasn't provided.",
+            exception.DBInvalidUnicodeParameter().message
+        )
+
     def test_paginate_query_attribute_error(self):
         sqlalchemy.asc(self.model.user_id).AndReturn('asc')
         self.query.order_by('asc').AndReturn(self.query)
@@ -151,6 +232,11 @@ class TestPaginateQuery(test_base.BaseTestCase):
                           utils.paginate_query, self.query,
                           self.model, 5, ['foo'])
 
+    def test_paginate_query_attribute_error_invalid_sortkey_3(self):
+        self.assertRaises(exception.InvalidSortKey,
+                          utils.paginate_query, self.query,
+                          self.model, 5, ['asc-nullinvalid'])
+
     def test_paginate_query_assertion_error(self):
         self.mox.ReplayAll()
         self.assertRaises(AssertionError,
@@ -185,6 +271,36 @@ class TestPaginateQuery(test_base.BaseTestCase):
                              marker=self.marker,
                              sort_dirs=['asc', 'desc'])
 
+    def test_paginate_query_null(self):
+        self.mox.StubOutWithMock(self.model.user_id, 'isnot')
+        self.model.user_id.isnot(None).AndReturn('asc_null_1')
+        sqlalchemy.desc('asc_null_1').AndReturn('asc_null_2')
+        self.query.order_by('asc_null_2').AndReturn(self.query)
+
+        sqlalchemy.asc(self.model.user_id).AndReturn('asc_1')
+        self.query.order_by('asc_1').AndReturn(self.query)
+
+        self.mox.StubOutWithMock(self.model.project_id, 'is_')
+        self.model.project_id.is_(None).AndReturn('desc_null_1')
+        sqlalchemy.desc('desc_null_1').AndReturn('desc_null_2')
+        self.query.order_by('desc_null_2').AndReturn(self.query)
+
+        sqlalchemy.desc(self.model.project_id).AndReturn('desc_1')
+        self.query.order_by('desc_1').AndReturn(self.query)
+
+        self.mox.StubOutWithMock(sqlalchemy.sql, 'and_')
+        sqlalchemy.sql.and_(mock.ANY).AndReturn('some_crit')
+        sqlalchemy.sql.and_(mock.ANY, mock.ANY).AndReturn('another_crit')
+        self.mox.StubOutWithMock(sqlalchemy.sql, 'or_')
+        sqlalchemy.sql.or_('some_crit', 'another_crit').AndReturn('some_f')
+        self.query.filter('some_f').AndReturn(self.query)
+        self.query.limit(5).AndReturn(self.query)
+        self.mox.ReplayAll()
+        utils.paginate_query(self.query, self.model, 5,
+                             ['user_id', 'project_id'],
+                             marker=self.marker,
+                             sort_dirs=['asc-nullslast', 'desc-nullsfirst'])
+
     def test_paginate_query_value_error(self):
         sqlalchemy.asc(self.model.user_id).AndReturn('asc_1')
         self.query.order_by('asc_1').AndReturn(self.query)
@@ -193,6 +309,136 @@ class TestPaginateQuery(test_base.BaseTestCase):
                           self.query, self.model, 5, ['user_id', 'project_id'],
                           marker=self.marker, sort_dirs=['asc', 'mixed'])
 
+    def test_paginate_on_hybrid(self):
+        sqlalchemy.asc(self.model.user_id).AndReturn('asc_1')
+        self.query.order_by('asc_1').AndReturn(self.query)
+
+        sqlalchemy.desc(self.model.some_hybrid).AndReturn('desc_1')
+        self.query.order_by('desc_1').AndReturn(self.query)
+
+        self.query.limit(5).AndReturn(self.query)
+        self.mox.ReplayAll()
+        utils.paginate_query(self.query, self.model, 5,
+                             ['user_id', 'some_hybrid'],
+                             sort_dirs=['asc', 'desc'])
+
+
+class Test_UnstableSortingOrder(test_base.BaseTestCase):
+    def test_multiple_primary_keys_stable(self):
+        self.assertTrue(
+            utils._stable_sorting_order(
+                FakeTableWithMultipleKeys, ['key1', 'key2']))
+
+    def test_classically_mapped_primary_keys_stable(self):
+        self.assertTrue(
+            utils._stable_sorting_order(FakeTableClassicalyMapped, ['id']))
+
+    def test_multiple_primary_keys_unstable(self):
+        self.assertFalse(
+            utils._stable_sorting_order(
+                FakeTableWithMultipleKeys, ['key1', 'key3']))
+
+    def test_joined_inh_stable(self):
+        self.assertTrue(
+            utils._stable_sorting_order(FakeTableJoinedInh, ['user_id'])
+        )
+
+    def test_single_inh_stable(self):
+        self.assertTrue(
+            utils._stable_sorting_order(FakeTableSingleInh, ['user_id'])
+        )
+
+    def test_unknown_primary_keys_stable(self):
+        self.assertIsNone(
+            utils._stable_sorting_order(object, ['key1', 'key2']))
+
+    def test_unique_index_stable(self):
+        self.assertTrue(
+            utils._stable_sorting_order(
+                FakeTableWithIndexes, ['key1', 'key2']))
+
+    def test_unique_index_unstable(self):
+        self.assertFalse(
+            utils._stable_sorting_order(
+                FakeTableWithIndexes, ['key1', 'key3']))
+
+
+class TestGetUniqueKeys(test_base.BaseTestCase):
+    def test_multiple_primary_keys(self):
+        self.assertEqual(
+            [{'key1', 'key2'}],
+            utils.get_unique_keys(FakeTableWithMultipleKeys))
+
+    def test_unique_index(self):
+        self.assertEqual(
+            [{'id'}, {'key1', 'key2'}],
+            utils.get_unique_keys(FakeTableWithIndexes))
+
+    def test_unknown_primary_keys(self):
+        self.assertIsNone(utils.get_unique_keys(object))
+
+    def test_cache(self):
+
+        class CacheTable(object):
+            info = {}
+            constraints_called = 0
+            indexes_called = 0
+
+            @property
+            def constraints(self):
+                self.constraints_called += 1
+                return []
+
+            @property
+            def indexes(self):
+                self.indexes_called += 1
+                return []
+
+        class CacheModel(object):
+            pass
+
+        table = CacheTable()
+        mapper_mock = mock.Mock(mapped_table=table, local_table=table)
+        mapper_mock.base_mapper = mapper_mock
+        mock_inspect = mock.Mock(
+            return_value=mapper_mock)
+        model = CacheModel()
+        self.assertNotIn('oslodb_unique_keys', CacheTable.info)
+        with mock.patch("oslo_db.sqlalchemy.utils.inspect", mock_inspect):
+            utils.get_unique_keys(model)
+
+        self.assertIn('oslodb_unique_keys', CacheTable.info)
+        self.assertEqual(1, table.constraints_called)
+        self.assertEqual(1, table.indexes_called)
+
+        for i in range(10):
+            utils.get_unique_keys(model)
+
+        self.assertEqual(1, table.constraints_called)
+        self.assertEqual(1, table.indexes_called)
+
+
+class TestPaginateQueryActualSQL(test_base.BaseTestCase):
+
+    def test_paginate_on_hybrid_assert_stmt(self):
+        s = Session()
+        q = s.query(FakeTable)
+        q = utils.paginate_query(
+            q, FakeTable, 5,
+            ['user_id', 'some_hybrid'],
+            sort_dirs=['asc', 'desc'])
+        expected_core_sql = (
+            select([FakeTable]).
+            order_by(sqlalchemy.asc(FakeTable.user_id)).
+            order_by(sqlalchemy.desc(FakeTable.some_hybrid)).
+            limit(5)
+        )
+
+        self.assertEqual(
+            str(expected_core_sql.compile()),
+            str(q.statement.compile())
+        )
+
 
 class TestMigrationUtils(db_test_base.DbTestCase):
 
@@ -253,7 +499,7 @@ class TestMigrationUtils(db_test_base.DbTestCase):
         real_ids = [row[0] for row in
                     self.engine.execute(select([test_table.c.id])).fetchall()]
 
-        self.assertEqual(len(real_ids), len(expected_ids))
+        self.assertEqual(len(expected_ids), len(real_ids))
         for id_ in expected_ids:
             self.assertTrue(id_ in real_ids)
 
@@ -293,7 +539,7 @@ class TestMigrationUtils(db_test_base.DbTestCase):
         rows_select = base_select.where(table.c.deleted != table.c.id)
         row_ids = [row['id'] for row in
                    self.engine.execute(rows_select).fetchall()]
-        self.assertEqual(len(row_ids), len(expected_values))
+        self.assertEqual(len(expected_values), len(row_ids))
         for value in expected_values:
             self.assertTrue(value['id'] in row_ids)
 
@@ -302,8 +548,8 @@ class TestMigrationUtils(db_test_base.DbTestCase):
         deleted_rows_ids = [row['id'] for row in
                             self.engine.execute(
                                 deleted_rows_select).fetchall()]
-        self.assertEqual(len(deleted_rows_ids),
-                         len(values) - len(row_ids))
+        self.assertEqual(len(values) - len(row_ids),
+                         len(deleted_rows_ids))
         for value in soft_deleted_values:
             self.assertTrue(value['id'] in deleted_rows_ids)
 
@@ -331,12 +577,12 @@ class TestMigrationUtils(db_test_base.DbTestCase):
 
         insp = reflection.Inspector.from_engine(self.engine)
         real_indexes = insp.get_indexes(table_name)
-        self.assertEqual(len(real_indexes), 3)
+        self.assertEqual(3, len(real_indexes))
         for index in real_indexes:
             name = index['name']
             self.assertIn(name, indexes)
-            self.assertEqual(set(index['column_names']),
-                             set(indexes[name]))
+            self.assertEqual(set(indexes[name]),
+                             set(index['column_names']))
 
     def test_change_deleted_column_type_to_id_type_integer(self):
         table_name = 'abc'
@@ -380,11 +626,7 @@ class TestMigrationUtils(db_test_base.DbTestCase):
                                                     foo=fooColumn)
 
         table = utils.get_table(self.engine, table_name)
-        # NOTE(boris-42): There is no way to check has foo type CustomType.
-        #                 but sqlalchemy will set it to NullType. This has
-        #                 been fixed upstream in recent SA versions
-        if SA_VERSION < (0, 9, 0):
-            self.assertTrue(isinstance(table.c.foo.type, NullType))
+
         self.assertTrue(isinstance(table.c.deleted.type, Integer))
 
     def test_change_deleted_column_type_to_boolean(self):
@@ -435,12 +677,6 @@ class TestMigrationUtils(db_test_base.DbTestCase):
                       Column('deleted', Integer))
         table.create()
 
-        # reflection of custom types has been fixed upstream
-        if SA_VERSION < (0, 9, 0):
-            self.assertRaises(exception.ColumnError,
-                              utils.change_deleted_column_type_to_boolean,
-                              self.engine, table_name)
-
         fooColumn = Column('foo', CustomType())
         utils.change_deleted_column_type_to_boolean(self.engine, table_name,
                                                     foo=fooColumn)
@@ -453,6 +689,23 @@ class TestMigrationUtils(db_test_base.DbTestCase):
             self.assertTrue(isinstance(table.c.foo.type, NullType))
         self.assertTrue(isinstance(table.c.deleted.type, Boolean))
 
+    def test_detect_boolean_deleted_constraint_detection(self):
+        table_name = 'abc'
+        table = Table(table_name, self.meta,
+                      Column('id', Integer, primary_key=True),
+                      Column('deleted', Boolean))
+        ck = [
+            const for const in table.constraints if
+            isinstance(const, CheckConstraint)][0]
+
+        self.assertTrue(utils._is_deleted_column_constraint(ck))
+
+        self.assertFalse(
+            utils._is_deleted_column_constraint(
+                CheckConstraint("deleted > 5")
+            )
+        )
+
     @db_test_base.backend_specific('sqlite')
     def test_change_deleted_column_type_sqlite_drops_check_constraint(self):
         table_name = 'abc'
@@ -500,13 +753,86 @@ class TestMigrationUtils(db_test_base.DbTestCase):
                                                   query_insert)
         result_insert = self.conn.execute(insert_statement)
         # Verify we insert 4 rows
-        self.assertEqual(result_insert.rowcount, 4)
+        self.assertEqual(4, result_insert.rowcount)
 
         query_all = select([insert_table]).where(
             insert_table.c.uuid.in_(uuidstrs))
         rows = self.conn.execute(query_all).fetchall()
         # Verify we really have 4 rows in insert_table
-        self.assertEqual(len(rows), 4)
+        self.assertEqual(4, len(rows))
+
+    def test_insert_from_select_with_specified_columns(self):
+        insert_table_name = "__test_insert_to_table__"
+        select_table_name = "__test_select_from_table__"
+        uuidstrs = []
+        for unused in range(10):
+            uuidstrs.append(uuid.uuid4().hex)
+        insert_table = Table(
+            insert_table_name, self.meta,
+            Column('id', Integer, primary_key=True,
+                   nullable=False, autoincrement=True),
+            Column('uuid', String(36), nullable=False))
+        select_table = Table(
+            select_table_name, self.meta,
+            Column('id', Integer, primary_key=True,
+                   nullable=False, autoincrement=True),
+            Column('uuid', String(36), nullable=False))
+
+        insert_table.create()
+        select_table.create()
+        # Add 10 rows to select_table
+        for uuidstr in uuidstrs:
+            ins_stmt = select_table.insert().values(uuid=uuidstr)
+            self.conn.execute(ins_stmt)
+
+        # Select 4 rows in one chunk from select_table
+        column = select_table.c.id
+        query_insert = select([select_table],
+                              select_table.c.id < 5).order_by(column)
+        insert_statement = utils.InsertFromSelect(insert_table,
+                                                  query_insert, ['id', 'uuid'])
+        result_insert = self.conn.execute(insert_statement)
+        # Verify we insert 4 rows
+        self.assertEqual(4, result_insert.rowcount)
+
+        query_all = select([insert_table]).where(
+            insert_table.c.uuid.in_(uuidstrs))
+        rows = self.conn.execute(query_all).fetchall()
+        # Verify we really have 4 rows in insert_table
+        self.assertEqual(4, len(rows))
+
+    def test_insert_from_select_with_specified_columns_negative(self):
+        insert_table_name = "__test_insert_to_table__"
+        select_table_name = "__test_select_from_table__"
+        uuidstrs = []
+        for unused in range(10):
+            uuidstrs.append(uuid.uuid4().hex)
+        insert_table = Table(
+            insert_table_name, self.meta,
+            Column('id', Integer, primary_key=True,
+                   nullable=False, autoincrement=True),
+            Column('uuid', String(36), nullable=False))
+        select_table = Table(
+            select_table_name, self.meta,
+            Column('id', Integer, primary_key=True,
+                   nullable=False, autoincrement=True),
+            Column('uuid', String(36), nullable=False))
+
+        insert_table.create()
+        select_table.create()
+        # Add 10 rows to select_table
+        for uuidstr in uuidstrs:
+            ins_stmt = select_table.insert().values(uuid=uuidstr)
+            self.conn.execute(ins_stmt)
+
+        # Select 4 rows in one chunk from select_table
+        column = select_table.c.id
+        query_insert = select([select_table],
+                              select_table.c.id < 5).order_by(column)
+        insert_statement = utils.InsertFromSelect(insert_table,
+                                                  query_insert, ['uuid', 'id'])
+        self.assertRaises(exception.DBError, self.conn.execute,
+                          insert_statement)
 
 
 class PostgesqlTestMigrations(TestMigrationUtils,
@@ -535,14 +861,30 @@ class TestConnectionUtils(test_utils.BaseTestCase):
 
         self.connect_string = 'postgresql://dude:pass@localhost/test'
 
+        # NOTE(rpodolyaka): mock the dialect parts, so that we don't depend
+        # on psycopg2 (or any other DBAPI implementation) in these tests
+
+        @classmethod
+        def fake_dbapi(cls):
+            return mock.MagicMock()
+        patch_dbapi = mock.patch.object(psycopg2.PGDialect_psycopg2, 'dbapi',
+                                        new=fake_dbapi)
+        patch_dbapi.start()
+        self.addCleanup(patch_dbapi.stop)
+
+        patch_onconnect = mock.patch.object(psycopg2.PGDialect_psycopg2,
+                                            'on_connect')
+        patch_onconnect.start()
+        self.addCleanup(patch_onconnect.stop)
+
     def test_connect_string(self):
         connect_string = utils.get_connect_string(**self.full_credentials)
-        self.assertEqual(connect_string, self.connect_string)
+        self.assertEqual(self.connect_string, connect_string)
 
     def test_connect_string_sqlite(self):
         sqlite_credentials = {'backend': 'sqlite', 'database': 'test.db'}
         connect_string = utils.get_connect_string(**sqlite_credentials)
-        self.assertEqual(connect_string, 'sqlite:///test.db')
+        self.assertEqual('sqlite:///test.db', connect_string)
 
     def test_is_backend_avail(self):
         self.mox.StubOutWithMock(sqlalchemy.engine.base.Engine, 'connect')
@@ -585,7 +927,9 @@ class TestConnectionUtils(test_utils.BaseTestCase):
             exception.BackendNotAvailable,
             provision.Backend._ensure_backend_available, self.connect_string
         )
-        self.assertEqual("Could not connect", str(exc))
+        self.assertEqual(
+            "Backend 'postgresql' is unavailable: "
+            "Could not connect", str(exc))
         self.assertEqual(
             "The postgresql backend is unavailable: %s" % err,
             log.output.strip())
@@ -602,20 +946,22 @@ class TestConnectionUtils(test_utils.BaseTestCase):
             exception.BackendNotAvailable,
             provision.Backend._ensure_backend_available, self.connect_string
         )
-        self.assertEqual("No DBAPI installed", str(exc))
+        self.assertEqual(
+            "Backend 'postgresql' is unavailable: "
+            "No DBAPI installed", str(exc))
         self.assertEqual(
             "The postgresql backend is unavailable: Can't import "
             "DBAPI module foobar", log.output.strip())
 
     def test_get_db_connection_info(self):
         conn_pieces = parse.urlparse(self.connect_string)
-        self.assertEqual(utils.get_db_connection_info(conn_pieces),
-                         ('dude', 'pass', 'test', 'localhost'))
+        self.assertEqual(('dude', 'pass', 'test', 'localhost'),
+                         utils.get_db_connection_info(conn_pieces))
 
     def test_connect_string_host(self):
         self.full_credentials['host'] = 'myhost'
         connect_string = utils.get_connect_string(**self.full_credentials)
-        self.assertEqual(connect_string, 'postgresql://dude:pass@myhost/test')
+        self.assertEqual('postgresql://dude:pass@myhost/test', connect_string)
 
 
 class MyModelSoftDeletedProjectId(declarative_base(), models.ModelBase,
@@ -658,8 +1004,8 @@ class TestModelQuery(test_base.BaseTestCase):
             MyModelSoftDeleted, session=self.session, deleted=False)
 
         deleted_filter = mock_query.filter.call_args[0][0]
-        self.assertEqual(str(deleted_filter),
-                         'soft_deleted_test_model.deleted = :deleted_1')
+        self.assertEqual('soft_deleted_test_model.deleted = :deleted_1',
+                         str(deleted_filter))
         self.assertEqual(deleted_filter.right.value,
                          MyModelSoftDeleted.__mapper__.c.deleted.default.arg)
 
@@ -676,7 +1022,7 @@ class TestModelQuery(test_base.BaseTestCase):
     @mock.patch.object(utils, "_read_deleted_filter")
     def test_no_deleted_value(self, _read_deleted_filter):
         utils.model_query(MyModelSoftDeleted, session=self.session)
-        self.assertEqual(_read_deleted_filter.call_count, 0)
+        self.assertEqual(0, _read_deleted_filter.call_count)
 
     def test_project_filter(self):
         project_id = 10
@@ -687,9 +1033,9 @@ class TestModelQuery(test_base.BaseTestCase):
 
         deleted_filter = mock_query.filter.call_args[0][0]
         self.assertEqual(
-            str(deleted_filter),
-            'soft_deleted_project_id_test_model.project_id = :project_id_1')
-        self.assertEqual(deleted_filter.right.value, project_id)
+            'soft_deleted_project_id_test_model.project_id = :project_id_1',
+            str(deleted_filter))
+        self.assertEqual(project_id, deleted_filter.right.value)
 
     def test_project_filter_wrong_model(self):
         self.assertRaises(ValueError, utils.model_query,
@@ -702,9 +1048,9 @@ class TestModelQuery(test_base.BaseTestCase):
             session=self.session, project_id=(10, None))
 
         self.assertEqual(
-            str(mock_query.filter.call_args[0][0]),
             'soft_deleted_project_id_test_model.project_id'
-            ' IN (:project_id_1, NULL)'
+            ' IN (:project_id_1, NULL)',
+            str(mock_query.filter.call_args[0][0])
         )
 
     def test_model_query_common(self):
@@ -789,8 +1135,10 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
         dispatcher = orig = utils.dispatch_for_dialect("*")(
             callable_fn.default)
         dispatcher = dispatcher.dispatch_for("sqlite")(callable_fn.sqlite)
-        dispatcher = dispatcher.dispatch_for("mysql+mysqldb")(
-            callable_fn.mysql_mysqldb)
+        dispatcher = dispatcher.dispatch_for("mysql+pymysql")(
+            callable_fn.mysql_pymysql)
+        dispatcher = dispatcher.dispatch_for("mysql")(
+            callable_fn.mysql)
         dispatcher = dispatcher.dispatch_for("postgresql")(
             callable_fn.postgresql)
 
@@ -804,7 +1152,8 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
         for targ in [
             callable_fn.default,
             callable_fn.sqlite,
-            callable_fn.mysql_mysqldb,
+            callable_fn.mysql,
+            callable_fn.mysql_pymysql,
             callable_fn.postgresql,
             callable_fn.postgresql_psycopg2,
             callable_fn.pyodbc
@@ -814,8 +1163,10 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
         dispatcher = orig = utils.dispatch_for_dialect("*", multiple=True)(
             callable_fn.default)
         dispatcher = dispatcher.dispatch_for("sqlite")(callable_fn.sqlite)
-        dispatcher = dispatcher.dispatch_for("mysql+mysqldb")(
-            callable_fn.mysql_mysqldb)
+        dispatcher = dispatcher.dispatch_for("mysql+pymysql")(
+            callable_fn.mysql_pymysql)
+        dispatcher = dispatcher.dispatch_for("mysql")(
+            callable_fn.mysql)
         dispatcher = dispatcher.dispatch_for("postgresql+*")(
             callable_fn.postgresql)
         dispatcher = dispatcher.dispatch_for("postgresql+psycopg2")(
@@ -832,15 +1183,17 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
         dispatcher, callable_fn = self._single_fixture()
         dispatcher("sqlite://", 1)
         dispatcher("postgresql+psycopg2://u:p@h/t", 2)
-        dispatcher("mysql://u:p@h/t", 3)
-        dispatcher("mysql+mysqlconnector://u:p@h/t", 4)
+        dispatcher("mysql+pymysql://u:p@h/t", 3)
+        dispatcher("mysql://u:p@h/t", 4)
+        dispatcher("mysql+mysqlconnector://u:p@h/t", 5)
 
         self.assertEqual(
             [
                 mock.call.sqlite('sqlite://', 1),
                 mock.call.postgresql("postgresql+psycopg2://u:p@h/t", 2),
-                mock.call.mysql_mysqldb("mysql://u:p@h/t", 3),
-                mock.call.default("mysql+mysqlconnector://u:p@h/t", 4)
+                mock.call.mysql_pymysql("mysql+pymysql://u:p@h/t", 3),
+                mock.call.mysql("mysql://u:p@h/t", 4),
+                mock.call.mysql("mysql+mysqlconnector://u:p@h/t", 5),
             ],
             callable_fn.mock_calls)
 
@@ -958,10 +1311,10 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
 
     def test_single_retval(self):
         dispatcher, callable_fn = self._single_fixture()
-        callable_fn.mysql_mysqldb.return_value = 5
+        callable_fn.mysql_pymysql.return_value = 5
 
         self.assertEqual(
-            dispatcher("mysql://u:p@h/t", 3), 5
+            5, dispatcher("mysql+pymysql://u:p@h/t", 3)
         )
 
     def test_engine(self):
@@ -974,14 +1327,25 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
             callable_fn.mock_calls
         )
 
-    def test_url(self):
+    def test_url_pymysql(self):
         url = sqlalchemy.engine.url.make_url(
-            "mysql+mysqldb://scott:tiger@localhost/test")
+            "mysql+pymysql://scott:tiger@localhost/test")
         dispatcher, callable_fn = self._single_fixture()
 
         dispatcher(url, 15)
         self.assertEqual(
-            [mock.call.mysql_mysqldb(url, 15)],
+            [mock.call.mysql_pymysql(url, 15)],
+            callable_fn.mock_calls
+        )
+
+    def test_url_mysql_generic(self):
+        url = sqlalchemy.engine.url.make_url(
+            "mysql://scott:tiger@localhost/test")
+        dispatcher, callable_fn = self._single_fixture()
+
+        dispatcher(url, 15)
+        self.assertEqual(
+            [mock.call.mysql(url, 15)],
             callable_fn.mock_calls
         )
 
@@ -1026,9 +1390,10 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
         dispatcher, callable_fn = self._multiple_fixture()
 
         dispatcher("postgresql+pyodbc://", 1)
-        dispatcher("mysql://", 2)
+        dispatcher("mysql+pymysql://", 2)
         dispatcher("ibm_db_sa+db2://", 3)
         dispatcher("postgresql+psycopg2://", 4)
+        dispatcher("postgresql://", 5)
 
         # TODO(zzzeek): there is a deterministic order here, but we might
         # want to tweak it, or maybe provide options.  default first?
@@ -1038,12 +1403,18 @@ class TestDialectFunctionDispatcher(test_base.BaseTestCase):
                 mock.call.postgresql('postgresql+pyodbc://', 1),
                 mock.call.pyodbc('postgresql+pyodbc://', 1),
                 mock.call.default('postgresql+pyodbc://', 1),
-                mock.call.mysql_mysqldb('mysql://', 2),
-                mock.call.default('mysql://', 2),
+                mock.call.mysql_pymysql('mysql+pymysql://', 2),
+                mock.call.mysql('mysql+pymysql://', 2),
+                mock.call.default('mysql+pymysql://', 2),
                 mock.call.default('ibm_db_sa+db2://', 3),
                 mock.call.postgresql_psycopg2('postgresql+psycopg2://', 4),
                 mock.call.postgresql('postgresql+psycopg2://', 4),
                 mock.call.default('postgresql+psycopg2://', 4),
+                # note this is called because we resolve the default
+                # DBAPI for the url
+                mock.call.postgresql_psycopg2('postgresql://', 5),
+                mock.call.postgresql('postgresql://', 5),
+                mock.call.default('postgresql://', 5),
             ],
             callable_fn.mock_calls
         )
